####
#
# autoHijack.py
#
# Download the latest Hijack kernel from the Hijack homepage and install it in the empeg
#
# Requires:
# You must have a Python installation to run this script.
# You must have a working Hijack kernel on your empeg. [Update to empeg occurs via Hijack ftp.]
#
#
# Author:                                   Support:
# Gary Noel Boone, PhD                      None, see Programming Forum on Unofficial empeg BBS
# Gary on the Unofficial empeg BBS          http://empeg.comms.net/php/postlist.php?Cat=&Board=hackers_prog
#
# Revision history:
# February 03, 2002     --  v2  --      added validation of new kernel after flashing
# February 02, 2002     --  v1  --      first working version
# February 02, 2002     --  v0  --      initial coding
#
# Development and testing:
# WinXP, ActiveState Python 2.1.212
#
# License:
# This software is released under the GNU Public License
#
# Warrantee:
# No warrantee or fitness for any purpose is expressed or implied. This is experimental, hobbyist code.
# Use at your own risk. 
false = 0
true  = 1
##################################################################################################################
##################################################################################################################
# Parameters. Set these for your configuration:

# GL_empeg_ip_str can be the ip address or the name listed in your hosts file
GL_empeg_ip_str = '192.168.123.135'

# The working directory is where your kernels are kept. New downloaded kernels are placed
# here. To recover prior kernels, an older kernel must be in this directory.
# This string can be an absolute path or relative to the location of this script
GL_working_directory = 'prior kernels'

# Do you have a Mark 1 empeg or a Mark 2 empeg? Uncomment one...
#GL_mk_str = 'mk1'
GL_mk_str = 'mk2'

# Recover last kernel flag. Set to true, then run script to return to the last succesfully flashed kernel.
# No download will occur; autoHijack will load the kernel given in the last_known_good
# database. Note that by "good," we mean only that the flash to the empeg completed succesfully.
# This recovery system is only intended to help reload the last kernel if the flash fails.
# For this to work, there must be a kernel in the working directory corersponding to the path/file
# named in the database.
#   If the last flash was successful, then setting GL_restore_last_kernel to true will just repeat
# that flash. Default false.
GL_restore_last_kernel = false

# Do you want to overwrite kernels? The default assumes that downloads work correctly. Therefore,
# we save time by not downloading kernels that we have already downloaded and flashed.
# Making GL_overwrite_downloaded_kernels true forces the current Hijack kernel on the web to be
# downloaded and flashed to the empeg. Default false
GL_overwrite_downloaded_kernels = false

# Do you want to check that the flashed kernel is identical to the file downloaded from the web?
# It doesn't take much time to check. Default true.
GB_validate_flashed_kernel = true

##################################################################################################################
##################################################################################################################
# other globals. You shouldn't need to change these.
GL_hijack_url = 'http://empeg-hijack.sourceforge.net'   
GL_database_filename = 'last_known_good.dat'


import sys, os, shelve, urllib, re, string
from ftplib import FTP

# This class is used to validate the buffers of the kernel as it is read from the empeg.
# accumulate() is called as a callback from ftp.retrbinary, adding the size of each read block
# to the running total
class accumulator:
    def __init__(self, orig_filename):
        self.valid = true
        self.total = 0
        try:
            self.orig_file = open(orig_filename, 'rb')
        except:
            print 'Unable to open original file for comparison: "' + orig_filename + '"'
            print 'Cannot validate new kernel. Exiting.'
            sys.exit(-1)

    # this function is called as each buffer is read from the kernel on the empeg.
    # Read a same-sized buffer from the file downloaded from the web. Check that
    # the two buffers are identical.
    # Note that the diskfile kernel is assumed to be shorter than the one downloaded
    # from the empeg. That's because the empeg always sends out more than the kernel
    # when you get /proc/empeg_kernel via ftp
    def accumulate(self, buffer_from_empeg):
        # only compare while there's more of the original file
        if self.orig_file:
            # read more of the kernel on disk
            buffer_from_file = self.orig_file.read(len(buffer_from_empeg))
            if not buffer_from_file:    # if eof
                self.orig_file.close()
                self.orig_file = None   # zero file object so we won't do any more
                return
            # Otherwise, we read some data from the file.
            # (The last good read from the file should be shorter than the buffer from the empeg,
            # so only compare up to that length.)
            self.total = self.total + len(buffer_from_file)     # accumulate size
            if buffer_from_file != buffer_from_empeg[:len(buffer_from_file)]:
                self.valid = false
                self.orig_file.close()
                self.orig_file = None   # zero file object so we won't do any more
                print '...Found differing buffers after ' + self.total
                print 'empeg buffer: ', buffer_from_empeg
                print 'file  buffer: ', buffer_from_file

    # return the total read size
    def get_total(self):
        return self.total

    # return the 'valid' instance variable's value    
    def is_valid(self):
        return self.valid
    

# return the name of the last succesfully flashed kernel. Its path/filename is stored in
# a shelve. We can't just use the highest version number in the kernel directory because we don't
# know how that file ended up there. It could have been downloaded separately by the user
# and never uploaded to the empeg.
def determine_last_good_kernel():
    database_fname = GL_working_directory + '/' + GL_database_filename
    if not os.path.exists(database_fname):
        return ''
    shv = shelve.open(database_fname)
    last_successful_path = shv[GL_empeg_ip_str]         # use the ip str as the dictionary key
    shv.close()
    print 'Last upload was "' + last_successful_path + '"'
    return last_successful_path


# store the path and filename of the last succesfully uploaded kernel
def update_last_known_good(uploaded_file_and_path):
    database_fname = GL_working_directory + '/' + GL_database_filename
    shv = shelve.open(database_fname)
    shv[GL_empeg_ip_str] = uploaded_file_and_path        # use the ip str as the dictionary key
    print 'Storing: last successful upload was "' + uploaded_file_and_path + '"'
    shv.close()


# open the Hijack web page and parse for the latest filename
# In its current version, the Hijack webpage just keeps the latest zImages.
# So we just have to find the one link that contains 'hijack', [mk1|mk2], and 'zImage'
# This routine will break if Mark Lord decides to keep multiple versions of hijack
# kernels on the Hijack webpage
def find_latest_file():
    try:
        page = urllib.urlopen(GL_hijack_url)
        html = page.read()
        page.close()
    except:
        print 'Error: Unable to read file "' + GL_hijack_url + '". Exiting.'
        sys.exit(-1)

    # find the links to hijack zImages in the page
    zimage_list = re.compile('<a href="(.*?hijack.*?zimage)"', re.IGNORECASE).findall(html)
    if not zimage_list:
        print 'Error: Could not find any Hijack kernels on "' + GL_hijack_url + '"'
        print 'No changes have been made to the empeg. Exiting.'
        sys.exit(-1)

    # now take just those images that are for the right Mark 1 or Mark 2 empeg
    hijack_list = [fn for fn in zimage_list if fn.lower().count(GL_mk_str)]
    if 1==len(hijack_list):
        return hijack_list[0]
    else:
        # Mark Lord has changed the format of his web page; we've found more than one hijack
        # kernel for our mk1 or mk2. There's more parsing to do to determine which file is the latest.
        print 'Error: Could not find the current Hijack kernel on "' + GL_hijack_url + '"'
        print 'No changes have been made to the empeg. Exiting.'
        sys.exit(-1)

# return the version number in the hijack filename: v169.hijack.v200b8.mk2.zimage --> 169
def version(fname):
    res_obj = re.compile('v(\d+)\D*', re.IGNORECASE).match(fname)
    if res_obj:
        return string.atoi(res_obj.group(1))
    else:
        return 0

# return the list of files in the working directory that are zImages, hijack kernels, and
# for the Mark 1 or Mark 2 indicated by GL_mk_str
def list_of_zImages():    
    files = os.listdir(GL_working_directory)
    files = [fn for fn in files if fn.lower().count('zimage')]
    files = [fn for fn in files if fn.lower().count('hijack')]
    files = [fn for fn in files if fn.lower().count(GL_mk_str)]
    return files

# sort function that takes two Hijack zImage filenames and compares their version numbers    
def sort_by_version(fn1, fn2):
    return version(fn2) - version(fn1)

# return true if the given zImage is a higher version than any of the kernels in the
# working directory
def newer_kernel_on_web(zimage_on_web):
    disk_zImages = list_of_zImages()        # read zImages in working directory
    disk_zImages.sort(sort_by_version)      # sort into descending order (highest version is in head of list)

    if version(zimage_on_web) > version(disk_zImages[0]):
        print 'New version available: "' + zimage_on_web + '"'
        return true
    else:
        print 'Version in "' + GL_working_directory + '" is current with Hijack website.'
        return false



# download the given zImagefilename into the local working directory
# return the path/filename where it was stored
def download_zImage(zimagefilename):
    print 'Downloading "' + zimagefilename + '"'
    zImage_url = GL_hijack_url + '/' + zimagefilename
    local_filename = GL_working_directory + '/' + zimagefilename
    try:
        urllib.urlretrieve(zImage_url, local_filename)
        return local_filename
    except:
        print 'Error: unable to download "' + zimagefilename + '" to "' + local_filename + '"'
        print 'Exiting.'
        sys.exit(-1)



# open the Hijack webpage, parse the html to find the latest Hijack zImage; download it
# return new kernel path/filename if we downloaded a file
def check_for_latest_file():
    zImage_filename = find_latest_file()
    if newer_kernel_on_web(zImage_filename) or GL_overwrite_downloaded_kernels:
        filepath = download_zImage(zImage_filename)
        return filepath
    else:
        return ''


# given the downloaded kernel and an ftp connection, check that they are the same
# Do this be downloading the new flashed kernel from the empeg, checking each 8k block
# as it arrives against the file downloaded from the web
def validate_kernel(ftp, filename):
    # Check validity of upload by downloading it to count the file size
    # [this method, used by mark when flashing was via /dev, does not work now that we flash to /proc. Hmmm ???]
    acc = accumulator(filename)                                 # file was closed after upload. acc will reopen it
    try:
        ftp.retrbinary('RETR /proc/empeg_kernel', acc.accumulate)
    except:
        print 'Error while downloading newly flashed kernel for verification. '
        print 'The empeg may now be in an usable state. You may with to try again with the last known good kernel.'
        print 'To do so, edit this script; set GL_restore_last_kernel to 1 at top. Then rerun script. Exiting.'
        sys.exit(-1)
    if acc.is_valid():
        print 'empeg kernel has been validated: ', acc.get_total(), ' bytes.'
    else:
        print 'empeg kernel differs from downloaded zImage file.', 
        print 'The empeg may now be in an usable state. You may with to try again with the last known good kernel.'
        print 'To do so, edit this script; set GL_restore_last_kernel to 1 at top. Then rerun script. Exiting.'
        sys.exit(-1)

    
# Using ftp, upgrade the empeg's kernel to the image given in 'filename'
def flash_kernel(filename):

    # check that the file exists
    if not os.path.exists(filename):
        print 'Error: File "' + filename + '" not found. Aborting upgrade. '
        print 'No changes have been made to empeg. Exiting.'
        sys.exit(-1)

    # open ftp, connect, prepare    
    ftp = FTP(GL_empeg_ip_str)   # connect to empeg
    ftp.login()               # user anonymous, passwd user@hostname
    ftp.set_pasv(0)

    # According to MLord, site commands return nothing but "okay." No need to error check them.

    # It would be good to have a way to pause the player before flashing. Unfortunately, we can't just     
    # send a play/pause to pause if playing . That will work if the empeg is playing, but if it's paused,
    # the empeg will start playing!
    #ftp.sendcmd('SITE button 20df16')      # send a play/pause
    
    # open the file and flash it
    new_kernel_file = open(filename, 'rb')                                      # open file
    ftp.sendcmd('SITE popup 99 Flashing new kernel...')                         # notify via empeg display
    try:
        response = ftp.storbinary('STOR /proc/empeg_kernel', new_kernel_file)   # flash
        print response
    except:
        print 'Error: There was an error while uploading. Here is the error info:'
        (exc_type, exc_value, exc_traceback) = sys.exc_info()
        print exc_type
        print exc_value
        print exc_traceback
        print 'Error while flashing kernel via ftp. '
        print 'The empeg may now be in an usable state. You may with to try again with the last known good kernel.'
        print 'To do so, edit this script; set GL_restore_last_kernel to 1 at top. Then rerun script. Exiting.'
        sys.exit(-1)
        
    ftp.sendcmd('SITE popup 1 Done')                                            # notify
    new_kernel_file.close()                                                     # close file

    if GB_validate_flashed_kernel:
        ftp.sendcmd('SITE popup 99 Validating new kernel...')                   # notify
        validate_kernel(ftp, filename)
        ftp.sendcmd('SITE popup 1 Done')                                        # notify

        
    # reboot the empeg to start the new kernel
    print 'Rebooting empeg player.'
    ftp.sendcmd('SITE reboot')

    # rebooting is likely to have a deleterious effect on the ftp connection, so no need to quit it.    
    #ftp.quit()


    
#
# main routine
# 
if __name__ == "__main__":

    # initialize last known good kernel path    
    last_kernel_path = determine_last_good_kernel()

    # restore?
    if GL_restore_last_kernel and last_kernel_path:
        print 'Reflashing "' + last_kernel_path + '"'
        flash_kernel(last_kernel_path)
        print 'Successfully reflashed kernel: ' + last_kernel_path
        print 'If you wish to try to upgrade again, Edit this script; set GL_restore_last_kernel to 0 at top.'
        print 'Then rerun script. Exiting.'
        sys.exit(0)


    # check website for latest Hijack version and download
    new_kernel_file = check_for_latest_file()
    if new_kernel_file:
        print 'Flashing "' + new_kernel_file + "'"
        flash_kernel(new_kernel_file)
        update_last_known_good(new_kernel_file)
        
    